/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarDeployableMetadataFactory.java 3852 2008-09-02 09:13:28Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.ejbjar.impl;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.EJBDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.common.impl.AbsDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.impl.xml.parsing.ParsingException;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.ejbjar.impl.configurator.EjbJarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.ejbjar.impl.specific.ISpecificEjbJarDeployableFactory;
import org.ow2.util.ee.metadata.ejbjar.impl.struct.JApplicationException;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.helper.EjbJarMetadataMerge;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.parsing.EJB3DeploymentDesc;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.AssemblyDescriptor;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.EJB3;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.EnterpriseBeans;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.Interceptor;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.Interceptors;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.MessageDriven;
import org.ow2.util.ee.metadata.ejbjar.impl.xml.struct.Session;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.IScanner;
import org.ow2.util.scan.api.classlocator.ArchiveClassesLocator;
import org.ow2.util.scan.api.classlocator.ClassLoaderClassesLocator;

/**
 * Factory to contruct ejb jar deployable metadata.
 * @param <E> ejb-jar type (EJB3, EJB2.1)
 * @author Gael Lalire
 */
public class EjbJarDeployableMetadataFactory<E extends EJBDeployable<E>> extends AbsDeployableMetadataFactory implements
        IEjbJarDeployableMetadataFactory<E> {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(EjbJarDeployableMetadataFactory.class);

    /**
     * Default Constructor.
     */
    public EjbJarDeployableMetadataFactory() {
    }

    /**
     * Constructor a scanner.
     * @param scanner the scanner
     */
    public EjbJarDeployableMetadataFactory(final IScanner scanner) {
        super(scanner);
    }

    /**
     * @param ejb3Deployable an ejb3 deployable
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public IEjbJarDeployableMetadata<E> createDeployableMetadata(final E ejb3Deployable) throws DeployableMetadataException {
        return createDeployableMetadata(ejb3Deployable, null, null);
    }

    /**
     * @param deployable an ejb3 deployable
     * @param classLoader a classloader to use if a class is not in archive
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public IEjbJarDeployableMetadata<E> createDeployableMetadata(final E deployable, final ClassLoader classLoader)
            throws DeployableMetadataException {
        List<? extends IClassesLocator> classesLocatorList = null;
        if (classLoader != null) {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(classLoader));
        } else {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(ClassLoader.getSystemClassLoader()));
        }
        return createDeployableMetadata(deployable, classesLocatorList, null);
    }

    /**
     * Extract a class list from xml.
     * @param ejb3 representation of ejb-jar.xml
     * @param classNameSet a set to fill with classes name referenced in ejb-jar.xml
     */
    private void completeClassList(final EJB3 ejb3, final Set<String> classNameSet) {
        if (ejb3 != null) {

            // Get enterprise beans classes
            EnterpriseBeans enterpriseBeans = ejb3.getEnterpriseBeans();
            if (enterpriseBeans != null) {
                // get session beans
                List<Session> sessionList = enterpriseBeans.getSessionList();
                for (Session session : sessionList) {
                    // get ejb-class
                    addClassToSetIfNotNull(session.getEjbClass(), classNameSet);
                }
                List<MessageDriven> messageDrivenList = enterpriseBeans.getMessageDrivenList();
                for (MessageDriven messageDriven : messageDrivenList) {
                    // get ejb-class
                    addClassToSetIfNotNull(messageDriven.getEjbClass(), classNameSet);
                }
            }

            // Get interceptor classes
            Interceptors interceptors = ejb3.getInterceptors();
            if (interceptors != null) {
                List<Interceptor> interceptorList = interceptors.getInterceptorList();
                if (interceptorList != null) {
                    for (Interceptor interceptor : interceptorList) {
                        addClassToSetIfNotNull(interceptor.getInterceptorClass(), classNameSet);
                    }
                }
            }

            // Get application exception classes
            AssemblyDescriptor assemblyDescriptor = ejb3.getAssemblyDescriptor();
            if (assemblyDescriptor != null) {
                List<JApplicationException> applicationExceptionList = assemblyDescriptor.getApplicationExceptionList();
                if (applicationExceptionList != null) {
                    for (JApplicationException applicationException : applicationExceptionList) {
                        addClassToSetIfNotNull(applicationException.getClassName(), classNameSet);
                    }
                }
            }

        }

    }


    /**
     * @param ejb3Deployable an ejb3 deployable
     * @param ejbJarExternLibraries extern libraries
     * @param specificDeployableFactoryList specific factories
     * @return metadata constructed using ejb3Deployable
     * @throws DeployableMetadataException if metadata can not be created
     */
    public IEjbJarDeployableMetadata<E> createDeployableMetadata(final E ejb3Deployable,
            final List<? extends IClassesLocator> ejbJarExternLibraries,
            final List<? extends ISpecificEjbJarDeployableFactory<E>> specificDeployableFactoryList)
            throws DeployableMetadataException {

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        IArchive archive = ejb3Deployable.getArchive();

        // Get EJB3 DD.
        EJB3 ejb3;
        try {
            ejb3 = EJB3DeploymentDesc.getEjb3(archive);
        } catch (ParsingException e) {
            throw new DeployableMetadataException(e);
        }

        EjbJarDeployableMetadataConfigurator<E> ejbJarArchiveMetadataConfigurator = new EjbJarDeployableMetadataConfigurator<E>(
                ejb3Deployable);
        Set<String> classNameSet = new HashSet<String>();

        if (specificDeployableFactoryList != null) {
            for (ISpecificEjbJarDeployableFactory<E> specificEjbJarDeployableFactory : specificDeployableFactoryList) {
                ejbJarArchiveMetadataConfigurator.addSpecificMetadataConfigurator(specificEjbJarDeployableFactory
                        .beforeScan(ejb3Deployable, classNameSet));
            }
        }

        // if metadata is complete then filter class to scan
        if (ejb3 != null && ejb3.isMetadataComplete()) {
            // disable annotation parsing
            ejbJarArchiveMetadataConfigurator.setAnnotationParsingDesactived(true);

        } else {
            // enable annotation parsing because metadata descriptor is not
            // complete
            ejbJarArchiveMetadataConfigurator.setAnnotationParsingDesactived(false);

            Iterator<String> entryIterator;
            try {
                entryIterator = archive.getEntries();
            } catch (ArchiveException e) {
                throw new DeployableMetadataException(e);
            }
            while (entryIterator.hasNext()) {
                String entry = entryIterator.next();
                if (entry.endsWith(CLASS_SUFFIX)) {
                    classNameSet.add(entry.substring(0, entry.length() - CLASS_SUFFIX.length()));
                }
            }
        }

        // get class to scan from deployment descriptor
        completeClassList(ejb3, classNameSet);

        // do annotation scan
        scan(classNameSet, new ArchiveClassesLocator(archive), ejbJarExternLibraries, ejbJarArchiveMetadataConfigurator);

        // get scan result
        EjbJarDeployableMetadata<E> ejbJarArchiveMetadata = ejbJarArchiveMetadataConfigurator.getEjbJarArchiveMetadata();

        // prepare the metadata for merge
        ejbJarArchiveMetadata.setEjb3(ejb3);

        // Merge DD with annotations metadata
        EjbJarMetadataMerge.merge(ejbJarArchiveMetadata);


        if (specificDeployableFactoryList != null) {
            for (ISpecificEjbJarDeployableFactory<E> specificEjbJarDeployableFactory : specificDeployableFactoryList) {
                specificEjbJarDeployableFactory.afterScan();
            }
        }

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
        }

        return ejbJarArchiveMetadata;
    }

    /**
     * @return the class returned by creation of metadata :
     *         IEJB3DeployableMetadata.class
     */
    @SuppressWarnings("unchecked")
    public Class<IEjbJarDeployableMetadata<E>> getDeployableMetadataClass() {
        return (Class<IEjbJarDeployableMetadata<E>>) (Class<?>) IEjbJarDeployableMetadata.class;
    }

}
